#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <gnome.h>
#include "rfb.h"
#include "vnc-viewer.h"

#include "snap_small.xpm"
#include "snap_big.xpm"

static GnomeApp *app;
static GtkWidget *scrolledwin;
static VncViewer *vnc = NULL;

static gchar *curDisplay = NULL;
static gboolean curShared = FALSE;
static VncViewerDepthType curDepth = VNC_VIEWER_DEPTH_DEFAULT;

static void show_connect_dlg(void);
static void disconnect_session(void);
static void take_screenshot(void);
static void quit_app(void);
static void show_about_dlg(void);

#define DISCONNECT_ITEM 1
#define SCREENSHOT_ITEM 3
static GnomeUIInfo file_menu[] = {
  GNOMEUIINFO_ITEM_STOCK(N_("_Connect..."), N_("Connect to a VNC server"),
			 show_connect_dlg, GNOME_STOCK_MENU_JUMP_TO),
  GNOMEUIINFO_ITEM_STOCK(N_("_Disconnect"), N_("Disconnect current VNC session"),
			 disconnect_session, GNOME_STOCK_MENU_CLOSE),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM(N_("Take screenshot"), N_("Take a snapshot of the screen"),
		   take_screenshot, snap_small),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM_STOCK(N_("_Quit"), N_("Quit application"),
			 quit_app, GNOME_STOCK_MENU_QUIT),
  GNOMEUIINFO_END
};

static GnomeUIInfo help_menu[] = {
  GNOMEUIINFO_ITEM_STOCK(N_("_About"), N_("Show an about box"),
			 show_about_dlg, GNOME_STOCK_MENU_ABOUT),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_HELP(PACKAGE),
  GNOMEUIINFO_END
};

static GnomeUIInfo menus[] = {
  GNOMEUIINFO_SUBTREE(N_("_File"), file_menu),
  GNOMEUIINFO_SUBTREE(N_("_Help"), help_menu),
  GNOMEUIINFO_END
};

#define DISCONNECT_BUTTON 1
#define SCREENSHOT_BUTTON 3
static GnomeUIInfo toolbar[] = {
  GNOMEUIINFO_ITEM_STOCK(N_("Connect"), N_("Connect to a VNC server"),
			 show_connect_dlg, GNOME_STOCK_PIXMAP_JUMP_TO),
  GNOMEUIINFO_ITEM_STOCK(N_("Disconnect"), N_("Disconnect current VNC session"),
			 disconnect_session, GNOME_STOCK_PIXMAP_CLOSE),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM(N_("Screenshot"), N_("Take a snapshot of the screen"),
		   take_screenshot, snap_big),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM_STOCK(N_("Quit"), N_("Quit application"),
			 quit_app, GNOME_STOCK_PIXMAP_QUIT),
  GNOMEUIINFO_END
};

static void getString(gchar *string, gpointer data) {
  gchar **ret = data;

  *ret = g_strdup(string);
}

static void vnc_connection_closed(void) {
  gtk_container_remove(GTK_CONTAINER(scrolledwin), GTK_WIDGET(vnc));
  vnc = NULL;
  if (curDisplay)
    g_free(curDisplay);
  curDisplay = NULL;
  gtk_window_set_title(GTK_WINDOW(app), _("Gnome VNC Viewer"));

  gtk_widget_set_sensitive(file_menu[DISCONNECT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(toolbar[DISCONNECT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(file_menu[SCREENSHOT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(toolbar[SCREENSHOT_ITEM].widget, FALSE);

  gtk_window_set_policy(GTK_WINDOW(app), FALSE, TRUE, FALSE);
}

static void vnc_connection_died(void) {
  gnome_app_message(app, "Connection to server lost");

  vnc_connection_closed();
}

static void make_connection(gchar *display, gboolean shared,
			    VncViewerDepthType depth) {
  gint fd;
  gint32 res;
  gchar *msg, *tmp;
  GtkWidget *win;
  GdkGeometry geometry;
  GdkWindowHints geom_mask;

  fd = connectToServer(display);
  if (fd < 0 || (res = doVersionHandshake(fd)) < 0) {
    gnome_app_error(app, "Could not connect to server");
    if (fd >= 0) close(fd);
    return;
  }
  switch (res) {
  case rfbConnFailed:
    tmp = getFailReason(fd);
    msg = g_strconcat("Connection failed: ", tmp, NULL);
    g_free(tmp);
    gnome_app_error(app, msg);
    g_free(msg);
    close(fd);
    return;
  case rfbNoAuth:
    break;
  case rfbVncAuth:
    msg = g_strdup_printf("Enter password for %s", display);
    win = gnome_request_dialog(TRUE, msg, "", -1,
			       getString, &msg,
			       GTK_WINDOW(app));
    g_free(msg);
    msg = NULL;
    gnome_dialog_close_hides(GNOME_DIALOG(win), FALSE);
    gtk_widget_show(win);
    if (gnome_dialog_run_and_close(GNOME_DIALOG(win)) != GNOME_OK ||
	msg == NULL) {
      close(fd);
      return;
    }
    switch (performAuth(fd, msg)) {
    case rfbVncAuthOK:
      g_free(msg);
      break;
    case rfbVncAuthFailed:
      g_free(msg);
      gnome_app_error(app, "Authentication to VNC server failed");
      close(fd);
      return;
    case rfbVncAuthTooMany:
      g_free(msg);
      gnome_app_error(app, "Too many authentication failures occured");
      close(fd);
      return;
    }
    break;
  }

  vnc = VNC_VIEWER(vnc_viewer_new(fd, shared, depth));
  vnc_viewer_set_cursor(vnc, VNC_VIEWER_CURSOR_DOT);
  gtk_signal_connect(GTK_OBJECT(vnc), "connection_died",
		     GTK_SIGNAL_FUNC(vnc_connection_died), NULL);
  gtk_container_add(GTK_CONTAINER(scrolledwin), GTK_WIDGET(vnc));
  gtk_widget_show(GTK_WIDGET(vnc));
  gtk_widget_grab_focus(GTK_WIDGET(vnc));

  geom_mask = GDK_HINT_MAX_SIZE;
  geometry.max_width = vnc->width;
  geometry.max_height = vnc->height;
  gtk_window_set_geometry_hints(GTK_WINDOW(app), GTK_WIDGET(scrolledwin),
				&geometry, geom_mask);

  msg = g_strconcat(_("Gnome VNC Viewer: "), vnc->desk_name, NULL);
  gtk_window_set_title(GTK_WINDOW(app), msg);
  g_free(msg);

  if (display != curDisplay) {
    if (curDisplay != NULL) g_free(curDisplay);
    curDisplay = g_strdup(display);
  }
  curShared = shared;
  curDepth = depth;

  gtk_widget_set_sensitive(file_menu[DISCONNECT_ITEM].widget, TRUE);
  gtk_widget_set_sensitive(toolbar[DISCONNECT_ITEM].widget, TRUE);
  gtk_widget_set_sensitive(file_menu[SCREENSHOT_ITEM].widget, TRUE);
  gtk_widget_set_sensitive(toolbar[SCREENSHOT_ITEM].widget, TRUE);

  gtk_window_set_policy(GTK_WINDOW(app), TRUE, FALSE, FALSE);
}



static void show_connect_dlg(void) {
  GtkWidget *win, *hbox, *wid, *hostname, *shared, *depth, *menu, *item;
  VncViewerDepthType dispDepth;
  
  win = gnome_dialog_new(_("Connect"), GNOME_STOCK_BUTTON_OK,
			 GNOME_STOCK_BUTTON_CLOSE, NULL);
  gnome_dialog_close_hides(GNOME_DIALOG(win), TRUE);
  gnome_dialog_set_parent(GNOME_DIALOG(win), GTK_WINDOW(app));

  hbox = gtk_hbox_new(FALSE, GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(win)->vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  wid = gtk_label_new(_("Host:"));
  gtk_box_pack_start(GTK_BOX(hbox), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);

  hostname = gnome_entry_new("past-hosts");
  gtk_box_pack_start(GTK_BOX(hbox), hostname, TRUE, TRUE, 0);
  gtk_widget_show(hostname);

  gnome_dialog_editable_enters(GNOME_DIALOG(win),
		GTK_EDITABLE(gnome_entry_gtk_entry(GNOME_ENTRY(hostname))));

  shared = gtk_check_button_new_with_label(_("Share this server with others"));
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(win)->vbox), shared, FALSE, TRUE, 0);
  gtk_widget_show(shared);

  hbox = gtk_hbox_new(FALSE, GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(win)->vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  wid = gtk_label_new(_("Display depth:"));
  gtk_box_pack_start(GTK_BOX(hbox), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);

  depth = gtk_option_menu_new();
  gtk_box_pack_start(GTK_BOX(hbox), depth, TRUE, TRUE, 0);
  gtk_widget_show(depth);

  menu = gtk_menu_new();
  item = gtk_menu_item_new_with_label("Default");
  gtk_object_set_data(GTK_OBJECT(item), "display-depth",
		      GINT_TO_POINTER(VNC_VIEWER_DEPTH_DEFAULT));
  gtk_container_add(GTK_CONTAINER(menu), item);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("8 bit");
  gtk_object_set_data(GTK_OBJECT(item), "display-depth",
		      GINT_TO_POINTER(VNC_VIEWER_DEPTH_8BIT));
  gtk_container_add(GTK_CONTAINER(menu), item);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("16 bit");
  gtk_object_set_data(GTK_OBJECT(item), "display-depth",
		      GINT_TO_POINTER(VNC_VIEWER_DEPTH_16BIT));
  gtk_container_add(GTK_CONTAINER(menu), item);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("24 bit");
  gtk_object_set_data(GTK_OBJECT(item), "display-depth",
		      GINT_TO_POINTER(VNC_VIEWER_DEPTH_24BIT));
  gtk_container_add(GTK_CONTAINER(menu), item);
  gtk_widget_show(item);

  gtk_option_menu_set_menu(GTK_OPTION_MENU(depth), menu);

  if (gnome_dialog_run_and_close(GNOME_DIALOG(win)) != 0) {
    gtk_widget_destroy(win);
    return;
  }

  if (vnc != NULL)
    vnc_connection_closed();
  dispDepth = GPOINTER_TO_INT(gtk_object_get_data(
	GTK_OBJECT(GTK_OPTION_MENU(depth)->menu_item), "display-depth"));
  make_connection(
	gtk_entry_get_text(GTK_ENTRY(gnome_entry_gtk_entry(
					GNOME_ENTRY(hostname)))),
	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(shared)),
	
	dispDepth);

  gtk_widget_destroy(win);
}

static void disconnect_session(void) {
  vnc_connection_closed();
}

static void save_screenshot(GtkFileSelection *win) {
  GdkImlibImage *img = gtk_object_get_data(GTK_OBJECT(win), "imlib-image");

  gdk_imlib_save_image(img,
		       gtk_entry_get_text(GTK_ENTRY(win->selection_entry)),
		       NULL);
  gtk_object_destroy(GTK_OBJECT(win));
}

static void take_screenshot(void) {
  GdkImlibImage *img;
  GtkWidget *win;

  g_return_if_fail(vnc != NULL);
  img = gdk_imlib_create_image_from_drawable(vnc->offscreen, NULL, 0, 0,
					     vnc->width, vnc->height);
  /* it would be pointless to save something like this in the image cache */
  gdk_imlib_changed_image(img);

  win = gtk_file_selection_new(_("Save Screenshot"));

  gtk_object_set_data_full(GTK_OBJECT(win), "imlib-image", img,
			   (GtkDestroyNotify)gdk_imlib_destroy_image);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(win)->ok_button),
			    "clicked", GTK_SIGNAL_FUNC(save_screenshot),
			    GTK_OBJECT(win));
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(win)->cancel_button),
			    "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
			    GTK_OBJECT(win));

  gtk_widget_show(win);
}

static void quit_app(void) {
  gtk_main_quit();
}

static void show_about_dlg(void) {
  GtkWidget *about;
  const gchar *authors[] = { "James Henstridge", NULL };

  about = gnome_about_new(_("Gnome VNC Viewer"), VERSION,
			  "Copyright (C) 1999  James Henstridge",
			  authors,
			  _("A virtual network computing viewer for GNOME"),
			  NULL);
  gtk_widget_show(about);
}

/* arg parsing stuff ... */
static void parseAnArg(poptContext con, enum poptCallbackReason reason,
		       const struct poptOption *opt,
		       const char *arg, void *data) {
  char *str;

  switch (reason) {
  case POPT_CALLBACK_REASON_OPTION:
    switch (opt->val) {
    case -1:
      curShared = TRUE; break;
    case -2:
      
    }
  case POPT_CALLBACK_REASON_PRE:
    if (!strcmp(arg, "8"))
      curDepth = VNC_VIEWER_DEPTH_8BIT;
    else if (!strcmp(arg, "16"))
      curDepth = VNC_VIEWER_DEPTH_16BIT;
    else if (!strcmp(arg, "24"))
      curDepth = VNC_VIEWER_DEPTH_24BIT;
    else if (!strcmp(arg, "default"))
      curDepth = VNC_VIEWER_DEPTH_DEFAULT;
    else {
      poptPrintUsage(con, stdout, 0);
      exit(1);
    }
    break;
  case POPT_CALLBACK_REASON_POST:
    str = poptGetArg(con);
    if (str) {
      curDisplay = g_strdup(str);
      if (poptGetArg(con) != NULL) {
	poptPrintUsage(con, stdout, 0);
	exit(1);
      }
    }
    break;
  }
}
static const struct poptOption vnc_options[] = {
  { NULL, '\0', POPT_ARG_CALLBACK|POPT_CBFLAG_POST,
    &parseAnArg, 0, NULL },
  { "shared", 's', POPT_ARG_NONE, NULL, -1,
    N_("share this display"), NULL },
  { "depth", 'd', POPT_ARG_STRING, NULL, -2,
    N_("the maximum bit depth for the display"), NULL },
  POPT_AUTOHELP
  { NULL, '\0', 0, NULL, 0 }
};

static int save_state(GnomeClient *client, gint phase,
		      GnomeRestartStyle save_style, gint shutdown,
		      GnomeInteractStyle interact_styyle, gint fast,
		      gchar *argv0) {
  char *argv[6];
  int argc;

  argc = 1;
  argv[0] = argv0;

  if (curDisplay) {
    if (curShared)
      argv[argc++] = "--shared";
    switch(curDepth) {
    case VNC_VIEWER_DEPTH_DEFAULT:
      break;
    case VNC_VIEWER_DEPTH_8BIT:
      argv[argc++] = "--depth"; argv[argc++] = "8"; break;
    case VNC_VIEWER_DEPTH_16BIT:
      argv[argc++] = "--depth"; argv[argc++] = "16"; break;
    case VNC_VIEWER_DEPTH_24BIT:
      argv[argc++] = "--depth"; argv[argc++] = "24"; break;
    }
    argv[argc++] = curDisplay;
  }
  argv[argc] = NULL;

  gnome_client_set_clone_command(client, argc, argv);
  gnome_client_set_restart_command(client, argc, argv);

  return TRUE;
}

int main(int argc, char *argv[]) {
  GtkWidget *win, *sw, *wid;
  GnomeDockItem *item;
  GnomeClient *client;
  int fd;
  gint32 res;
  gchar *title;

  bindtextdomain(PACKAGE, "/usr/share/locale");
  textdomain(PACKAGE);

  gnome_init_with_popt_table(PACKAGE, VERSION, argc, argv,
			     vnc_options, 0, NULL);

  /* stuff to handle session management */
  client = gnome_master_client();
  gtk_signal_connect(GTK_OBJECT(client), "save_yourself",
		     GTK_SIGNAL_FUNC(save_state), argv[0]);
  gtk_signal_connect(GTK_OBJECT(client), "die",
		     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

  gtk_widget_push_visual(gdk_rgb_get_visual());
  gtk_widget_push_colormap(gdk_rgb_get_cmap());

  win = gnome_app_new("gnome-vnc", _("Gnome VNC Viewer"));
  app = GNOME_APP(win);

  gnome_app_create_menus(app, menus);
  gnome_app_create_toolbar(app, toolbar);
  item = gnome_app_get_dock_item_by_name(app, GNOME_APP_TOOLBAR_NAME);
  wid = gnome_dock_item_get_child(item);
  gtk_toolbar_set_style(GTK_TOOLBAR(wid), GTK_TOOLBAR_ICONS);

  gtk_widget_set_sensitive(file_menu[DISCONNECT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(toolbar[DISCONNECT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(file_menu[SCREENSHOT_ITEM].widget, FALSE);
  gtk_widget_set_sensitive(toolbar[SCREENSHOT_ITEM].widget, FALSE);

  gtk_signal_connect(GTK_OBJECT(win), "destroy",
		     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

  gtk_window_set_default_size(GTK_WINDOW(win), 400, 400);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gnome_app_set_contents(app, sw);
  gtk_widget_show(sw);

  gtk_widget_show(win);

  scrolledwin = sw;

  if (curDisplay)
    make_connection(curDisplay, curShared, curDepth);

  gtk_main();
  return 0;
}
