#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <gtk/gtkmarshal.h>
#include <gdk_imlib.h>
#include "vnc-viewer.h"

#include <string.h>
#include <unistd.h>

enum {
  SET_SCROLL_ADJUSTMENTS,
  VNC_FRAMEBUFFER_UPDATE,
  VNC_SET_COLOURMAP_ENTRIES,
  VNC_BELL,
  VNC_SERVER_CUT_TEXT,
  CONNECTION_DIED,
  LAST_SIGNAL
};

static guint vnc_signals[LAST_SIGNAL] = { 0 };
static GtkWidgetClass *parent_class = NULL;

static void vnc_viewer_class_init(VncViewerClass *klass);
static void vnc_viewer_init(VncViewer *vnc);

GtkType vnc_viewer_get_type(void) {
  static GtkType vnc_type = 0;
  if (!vnc_type) {
    GtkTypeInfo vnc_info = {
      "VncViewer",
      sizeof(VncViewer),
      sizeof(VncViewerClass),
      (GtkClassInitFunc) vnc_viewer_class_init,
      (GtkObjectInitFunc) vnc_viewer_init,
      NULL,
      NULL,
    };
    vnc_type = gtk_type_unique(gtk_widget_get_type(), &vnc_info);
  }
  return vnc_type;
}

static void vnc_viewer_set_scroll_adjustments(VncViewer *vnc,
					      GtkAdjustment *hadj,
					      GtkAdjustment *vadj);
static void vnc_viewer_framebuffer_update(VncViewer *vnc,
					  rfbServerToClientMsg *msg);
static void vnc_viewer_set_colourmap_entries(VncViewer *vnc,
					     rfbServerToClientMsg *msg,
					     guint16 *entries);
static void vnc_viewer_bell(VncViewer *vnc, rfbServerToClientMsg *msg);
static void vnc_viewer_server_cut_text(VncViewer *vnc,
				       rfbServerToClientMsg *msg, gchar *text);

static void vnc_viewer_realize(GtkWidget *widget);
static void vnc_viewer_size_allocate(GtkWidget *widget,
				     GtkAllocation *allocation);
static gint vnc_viewer_expose_event(GtkWidget *widget, GdkEventExpose *event);
static gint vnc_viewer_button_press_event(GtkWidget *widget,
					  GdkEventButton *event);
static gint vnc_viewer_button_release_event(GtkWidget *widget,
					    GdkEventButton *event);
static gint vnc_viewer_motion_notify_event(GtkWidget *widget,
					   GdkEventMotion *event);
static gint vnc_viewer_key_press_event(GtkWidget *widget,
				       GdkEventKey *event);
static gint vnc_viewer_key_release_event(GtkWidget *widget,
					 GdkEventKey *event);

static void vnc_viewer_destroy(GtkObject *object);

static void vnc_viewer_class_init(VncViewerClass *klass) {
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = GTK_OBJECT_CLASS(klass);
  widget_class = GTK_WIDGET_CLASS(klass);
  parent_class = gtk_type_class(gtk_widget_get_type());

  vnc_signals[SET_SCROLL_ADJUSTMENTS] =
    gtk_signal_new("set_scroll_adjustments",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass, set_scroll_adjustments),
		   gtk_marshal_NONE__POINTER_POINTER,
		   GTK_TYPE_NONE, 2,
		   GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);
  widget_class->set_scroll_adjustments_signal =
    vnc_signals[SET_SCROLL_ADJUSTMENTS];
  vnc_signals[VNC_FRAMEBUFFER_UPDATE] =
    gtk_signal_new("vnc_framebuffer_update",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass, vnc_framebuffer_update),
		   gtk_marshal_NONE__POINTER,
		   GTK_TYPE_NONE, 1,
		   GTK_TYPE_POINTER);
  vnc_signals[VNC_SET_COLOURMAP_ENTRIES] =
    gtk_signal_new("vnc_set_colourmap_entries",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass,vnc_set_colourmap_entries),
		   gtk_marshal_NONE__POINTER_POINTER,
		   GTK_TYPE_NONE, 2,
		   GTK_TYPE_POINTER, GTK_TYPE_POINTER);
  vnc_signals[VNC_BELL] =
    gtk_signal_new("vnc_bell",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass, vnc_bell),
		   gtk_marshal_NONE__POINTER,
		   GTK_TYPE_NONE, 1,
		   GTK_TYPE_POINTER);
  vnc_signals[VNC_SERVER_CUT_TEXT] =
    gtk_signal_new("vnc_server_cut_text",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass, vnc_server_cut_text),
		   gtk_marshal_NONE__POINTER_POINTER,
		   GTK_TYPE_NONE, 2,
		   GTK_TYPE_POINTER, GTK_TYPE_STRING);
  vnc_signals[CONNECTION_DIED] =
    gtk_signal_new("connection_died",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(VncViewerClass, connection_died),
		   gtk_marshal_NONE__NONE,
		   GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals(object_class, vnc_signals, LAST_SIGNAL);

  klass->set_scroll_adjustments    = vnc_viewer_set_scroll_adjustments;
  klass->vnc_framebuffer_update    = vnc_viewer_framebuffer_update;
  klass->vnc_set_colourmap_entries = vnc_viewer_set_colourmap_entries;
  klass->vnc_bell                  = vnc_viewer_bell;
  klass->vnc_server_cut_text       = vnc_viewer_server_cut_text;

  /* signals to do with redrawing and the like */
  widget_class->realize = vnc_viewer_realize;
  widget_class->size_allocate = vnc_viewer_size_allocate;
  widget_class->expose_event = vnc_viewer_expose_event;

  /* signals that create events that should be sent to the server */
  widget_class->button_press_event = vnc_viewer_button_press_event;
  widget_class->button_release_event = vnc_viewer_button_release_event;
  widget_class->motion_notify_event = vnc_viewer_motion_notify_event;
  widget_class->key_press_event = vnc_viewer_key_press_event;
  widget_class->key_release_event = vnc_viewer_key_release_event;

  object_class->destroy = vnc_viewer_destroy;
}

static void vnc_viewer_init(VncViewer *self) {
  const char dotSrcBits[] = { 0, 14,14,14, 0 };
  const char dotMskBits[] = { 31,31,31,31,31 };
  const char noneBits[] = { 0 };
  GdkBitmap *src, *mask;
  GdkColor white, black;

  GTK_WIDGET_UNSET_FLAGS(self, GTK_NO_WINDOW);
  GTK_WIDGET_SET_FLAGS(self, GTK_CAN_FOCUS);

  self->cursor_type = VNC_VIEWER_CURSOR_NORMAL;

  self->fd = -1;
  self->hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 1, 1, 1, 1));
  self->vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 1, 1, 1, 1));

  gtk_signal_connect_object(GTK_OBJECT(self->hadj), "value_changed",
			    GTK_SIGNAL_FUNC(gtk_widget_queue_draw),
			    GTK_OBJECT(self));
  gtk_signal_connect_object(GTK_OBJECT(self->vadj), "value_changed",
			    GTK_SIGNAL_FUNC(gtk_widget_queue_draw),
			    GTK_OBJECT(self));

  self->in_tag = 0;
  self->out_tag = 0;

  self->out_queue = 0;

  self->width = 0;
  self->height = 0;
  memset(&(self->pixel_format), '\0', sizeof(self->pixel_format));
  self->desk_name = NULL;

  self->button_mask = 0;

  self->offscreen = NULL;

  gdk_color_white(gtk_widget_get_colormap(GTK_WIDGET(self)), &white);
  gdk_color_black(gtk_widget_get_colormap(GTK_WIDGET(self)), &black);

  src  = gdk_bitmap_create_from_data(NULL, dotSrcBits, 5, 5);
  mask = gdk_bitmap_create_from_data(NULL, dotMskBits, 5, 5);
  self->dot = gdk_cursor_new_from_pixmap(src, mask, &black, &white, 2, 2);
  gdk_bitmap_unref(src);
  gdk_bitmap_unref(mask);

  src = gdk_bitmap_create_from_data(NULL, noneBits, 1, 1);
  self->none = gdk_cursor_new_from_pixmap(src, src, &black, &white, 0, 0);
  gdk_bitmap_unref(src);
}

GtkWidget *vnc_viewer_new(gint fd, gboolean shared, VncViewerDepthType depth) {
  VncViewer *vnc = gtk_type_new(vnc_viewer_get_type());

  if (!vnc)
    return NULL;
  vnc_viewer_configure(vnc, fd, shared, depth);

  return GTK_WIDGET(vnc);
}

void vnc_viewer_close_connection(VncViewer *vnc) {
  /* if already closed, don't do it again */
  if (vnc->fd < 0) return;
  close(vnc->fd);
  vnc->fd = -1;
  if (vnc->in_tag)  gdk_input_remove(vnc->in_tag);
  if (vnc->out_tag) gdk_input_remove(vnc->out_tag);
  vnc->in_tag = vnc->out_tag = 0;
  g_list_foreach(vnc->out_queue, (GFunc)g_free, NULL);
  g_list_free(vnc->out_queue);
  vnc->out_queue = NULL;
}

void vnc_viewer_set_cursor(VncViewer *vnc, VncViewerCursorType cursor_type) {
  GdkCursor *cur;

  vnc->cursor_type = cursor_type;

  if (GTK_WIDGET_REALIZED(vnc))
    switch (vnc->cursor_type) {
    case VNC_VIEWER_CURSOR_NORMAL:
      cur = gdk_cursor_new(GDK_ARROW);
      gdk_window_set_cursor(GTK_WIDGET(vnc)->window, cur);
      gdk_cursor_destroy(cur);
      break;
    case VNC_VIEWER_CURSOR_DOT:
      gdk_window_set_cursor(GTK_WIDGET(vnc)->window, vnc->dot);
      break;
    case VNC_VIEWER_CURSOR_NONE:
      gdk_window_set_cursor(GTK_WIDGET(vnc)->window, vnc->none);
      break;
    }
}

static void vnc_viewer_get_event(VncViewer *vnc, gint fd,
				 GdkInputCondition cond);
static void vnc_viewer_send_event(VncViewer *vnc, gint fd,
				  GdkInputCondition cond);

static void vnc_viewer_queue_event(VncViewer *vnc, rfbClientToServerMsg *msg) {
  if (vnc->fd < 0) {
    g_free(msg);
    return;
  }

  vnc->out_queue = g_list_append(vnc->out_queue, msg);

  if (vnc->out_tag == 0)
    vnc->out_tag = gdk_input_add(vnc->fd, GDK_INPUT_WRITE,
				 (GdkInputFunction)vnc_viewer_send_event,
				 vnc);
}

/* fd is a connection to the VNC server just after the connection has been
 * authenticated.  Shared says whether we allow other concurrent connections,
 * and bpp is the desired bpp */
void vnc_viewer_configure(VncViewer *vnc, gint fd, gboolean shared,
			  VncViewerDepthType depth) {
  guint32 encodings[] = { 
    /*    rfbEncodingHextile, */
    rfbEncodingCoRRE, rfbEncodingRRE,
    rfbEncodingCopyRect, rfbEncodingRaw };
  guint16 nEncodings = sizeof(encodings) / sizeof(encodings[0]);
  guint32 localEncodings[] = { rfbEncodingCopyRect, rfbEncodingRaw };
  guint nLocalEncodings = sizeof(localEncodings) / sizeof(localEncodings[0]);

  vnc->fd = fd;

  if (sendClientInitMsg(fd, shared) < 0) {
    g_warning("client init message send failed");
    return;
  }
  if (getServerInitMsg(fd, &(vnc->width), &(vnc->height), &(vnc->pixel_format),
		       &(vnc->desk_name)) < 0) {
    g_warning("server init message recieve failed");
    return;
  }
  g_message("Desk Name='%s'", vnc->desk_name);
  g_message("Desk Size=%dx%d, Depth=%d", vnc->width, vnc->height,
	    (gint)vnc->pixel_format.depth);
  vnc->in_tag = gdk_input_add(vnc->fd, GDK_INPUT_READ,
			      (GdkInputFunction)vnc_viewer_get_event, vnc);

  vnc->pixel_format.bigEndian = TRUE;
  switch (depth) {
  case VNC_VIEWER_DEPTH_DEFAULT:
    break;
  case VNC_VIEWER_DEPTH_8BIT:
    if (vnc->pixel_format.depth > 8) {    /* only change depth down */
      vnc->pixel_format.bitsPerPixel = 8;
      vnc->pixel_format.depth = 8;
      vnc->pixel_format.trueColour = TRUE;
      vnc->pixel_format.redMax = 7;
      vnc->pixel_format.greenMax = 7;
      vnc->pixel_format.blueMax = 3;
      vnc->pixel_format.redShift = 0;
      vnc->pixel_format.greenShift = 3;
      vnc->pixel_format.blueShift = 6;
    }
    break;
  case VNC_VIEWER_DEPTH_16BIT:
    if (vnc->pixel_format.depth > 16) {    /* only change depth down */
      vnc->pixel_format.bitsPerPixel = 16;
      vnc->pixel_format.depth = 16;
      vnc->pixel_format.trueColour = TRUE;
      vnc->pixel_format.redMax = 31;
      vnc->pixel_format.greenMax = 63;
      vnc->pixel_format.blueMax = 31;
      vnc->pixel_format.redShift = 11;
      vnc->pixel_format.greenShift = 5;
      vnc->pixel_format.blueShift = 0;
    }
    break;
  case VNC_VIEWER_DEPTH_24BIT:
    if (vnc->pixel_format.depth > 24) {    /* only change depth down */
      vnc->pixel_format.bitsPerPixel = 32;
      vnc->pixel_format.depth = 24;
      vnc->pixel_format.trueColour = TRUE;
      vnc->pixel_format.redMax = 255;
      vnc->pixel_format.greenMax = 255;
      vnc->pixel_format.blueMax = 255;
      vnc->pixel_format.redShift = 16;
      vnc->pixel_format.greenShift = 8;
      vnc->pixel_format.blueShift = 0;
    }
    break;
  }
  vnc_viewer_queue_event(vnc, newSetPixelFormatMsg(&(vnc->pixel_format)));

  if (sameMachine(vnc->fd))
    vnc_viewer_queue_event(vnc, newSetEncodingsMsg(nLocalEncodings,
						   localEncodings));
  else
    vnc_viewer_queue_event(vnc, newSetEncodingsMsg(nEncodings, encodings));
}

static void vnc_viewer_get_event(VncViewer *vnc, gint fd,
				 GdkInputCondition cond) {
  rfbServerToClientMsg *msg;

  if (cond == GDK_INPUT_EXCEPTION) {
    vnc_viewer_close_connection(vnc);
    gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[CONNECTION_DIED]);
    return;
  }
  msg = getServerMsg(fd);

  if (!msg) {
    vnc_viewer_close_connection(vnc);
    gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[CONNECTION_DIED]);
    return;
  }

  switch (msg->type) {
  case rfbFramebufferUpdate:
    gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[VNC_FRAMEBUFFER_UPDATE], msg);
    break;
  case rfbSetColourMapEntries:
    {
      guint16 *entries = getColourMapEntries(fd, msg->scme.nColours);

      gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[VNC_SET_COLOURMAP_ENTRIES],
		      msg, entries);
      g_free(entries);
      break;
    }
  case rfbBell:
    gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[VNC_BELL], msg);
    break;
  case rfbServerCutText:
    {
      gchar *text = getServerCutText(fd, msg->sct.length);

      gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[VNC_SERVER_CUT_TEXT],
		      msg, text);
      g_free(text);
      break;
    }
  default:
  }
  g_free(msg);
}

static void vnc_viewer_send_event(VncViewer *vnc, gint fd,
				  GdkInputCondition cond) {
  rfbClientToServerMsg *msg;

  if (cond == GDK_INPUT_EXCEPTION) {
    vnc_viewer_close_connection(vnc);
    gtk_signal_emit(GTK_OBJECT(vnc), vnc_signals[CONNECTION_DIED]);
    return;
  }
  msg = vnc->out_queue->data;
  vnc->out_queue = g_list_remove(vnc->out_queue, msg);
  rfb_send_and_destroy_message(fd, msg);

  /* if no more queued events exist, remove this handler */
  if (vnc->out_queue == NULL) {
    gdk_input_remove(vnc->out_tag);
    vnc->out_tag = 0;
  }
}

static void vnc_viewer_set_scroll_adjustments(VncViewer *vnc,
					      GtkAdjustment *hadj,
					      GtkAdjustment *vadj) {
  if (vnc->hadj) {
    gtk_signal_disconnect_by_data(GTK_OBJECT(vnc->hadj), vnc);
    gtk_object_unref(GTK_OBJECT(vnc->hadj));
  }
  if (vnc->vadj) {
    gtk_signal_disconnect_by_data(GTK_OBJECT(vnc->vadj), vnc);
    gtk_object_unref(GTK_OBJECT(vnc->vadj));
  }
  /* assign new adjustment objects ... */
  vnc->hadj = hadj; vnc->vadj = vadj;
  if (hadj) gtk_object_ref(GTK_OBJECT(hadj));
  else      vnc->hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0,0,1,1,1,1));
  if (vadj) gtk_object_ref(GTK_OBJECT(vadj));
  else      vnc->vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0,0,1,1,1,1));

  vnc->hadj->value = 0;
  vnc->hadj->lower = 0;
  vnc->hadj->upper = vnc->width;
  vnc->hadj->step_increment = 20;
  vnc->hadj->page_increment = GTK_WIDGET(vnc)->requisition.width;
  vnc->hadj->page_size = GTK_WIDGET(vnc)->requisition.width;
  gtk_adjustment_changed(vnc->hadj);

  vnc->vadj->value = 0;
  vnc->vadj->lower = 0;
  vnc->vadj->upper = vnc->width;
  vnc->vadj->step_increment = 20;
  vnc->vadj->page_increment = GTK_WIDGET(vnc)->requisition.height;
  vnc->vadj->page_size = GTK_WIDGET(vnc)->requisition.height;
  gtk_adjustment_changed(vnc->vadj);

  gtk_signal_connect_object(GTK_OBJECT(vnc->hadj), "value_changed",
			    GTK_SIGNAL_FUNC(gtk_widget_queue_draw),
			    GTK_OBJECT(vnc));
  gtk_signal_connect_object(GTK_OBJECT(vnc->vadj), "value_changed",
			    GTK_SIGNAL_FUNC(gtk_widget_queue_draw),
			    GTK_OBJECT(vnc));
  gtk_widget_queue_draw(GTK_WIDGET(vnc));
}

static gboolean vnc_viewer_send_draw_request(VncViewer *vnc) {
  if (vnc->fd < 0) return FALSE;

  vnc_viewer_queue_event(vnc, newFramebufferUpdateRequestMsg(TRUE, 0, 0,
						vnc->width, vnc->height));
  return FALSE;
}

static void vnc_viewer_framebuffer_update(VncViewer *vnc,
					  rfbServerToClientMsg *msg) {
  guint32 rect, srect, tile, i, j;
  guchar *buffer;
  rfbPixelFormat *fmt = &(vnc->pixel_format);
  guint8 bpp = fmt->bitsPerPixel, tmpR, tmpG, tmpB, bgR, bgG, bgB;
  gdouble redScale = 255.0 / fmt->redMax, greenScale = 255.0 / fmt->greenMax,
    blueScale = 255.0 / fmt->blueMax;

  for (rect = 0; rect < msg->fu.nRects; rect++) {
    rfbRectangleDescription *desc =
      getNextRectangle(vnc->fd, vnc->pixel_format.bitsPerPixel/8);
    guint16 rWidth = desc->header.r.w, rHeight = desc->header.r.h;
    guint16 xPos, yPos;
    guint32 totPixels = rWidth * rHeight;

    //g_message("Rectangle (%d,%d) %dx%d Encoding=%d", (gint)desc->header.r.x,
    //	      (gint)desc->header.r.y, (gint)desc->header.r.w,
    //	      (gint)desc->header.r.h, (gint)desc->header.encoding);
    switch (desc->header.encoding) {
    case rfbEncodingRaw:
      buffer = g_malloc(totPixels * 3);
      for (i = 0; i < totPixels; i++) {
	guint32 pixel = 0;
	switch (bpp) {
	case 8:  pixel = desc->data.raw[i]; break;
	case 16: pixel = ((guint16 *)(desc->data.raw))[i]; break;
	case 32: pixel = ((guint32 *)(desc->data.raw))[i]; break;
	default: g_assert_not_reached();
	}
	buffer[3*i]   = (pixel >> fmt->redShift & fmt->redMax) * redScale;
	buffer[3*i+1] = (pixel >> fmt->greenShift & fmt->greenMax) *greenScale;
	buffer[3*i+2] = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
      }
      gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
			 desc->header.r.x, desc->header.r.y,
			 rWidth, rHeight,
			 GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
      g_free(buffer);
      break;
    case rfbEncodingCopyRect:
      gdk_draw_pixmap(vnc->offscreen, vnc->gc, vnc->offscreen,
		      desc->data.cr.srcX, desc->data.cr.srcY,
		      desc->header.r.x, desc->header.r.y,
		      rWidth, rHeight);
      break;
    case rfbEncodingRRE:
      buffer = g_malloc(totPixels * 3);
      tmpR = (desc->data.rre.bgPixel >> fmt->redShift & fmt->redMax) *
	redScale;
      tmpG = (desc->data.rre.bgPixel >> fmt->greenShift & fmt->greenMax) *
	greenScale;
      tmpB = (desc->data.rre.bgPixel >> fmt->blueShift & fmt->blueMax) *
	blueScale;

      for (i = 0; i < totPixels; i++) {
	buffer[3*i] = tmpR; buffer[3*i+1] = tmpG; buffer[3*i+2] = tmpB;
      }
      for (srect = 0; srect < desc->data.rre.nSubrects; srect++) {
	guint32 pixel = desc->data.rre.rects[srect].pixel;
	rfbRectangle *sr = &(desc->data.rre.rects[srect].rect);
	tmpR = (pixel >> fmt->redShift & fmt->redMax) * redScale;
	tmpG = (pixel >> fmt->greenShift & fmt->greenMax) * greenScale;
	tmpB = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
	for (j = sr->y; j < sr->y + sr->h; j++)
	  for (i = sr->x; i < sr->x + sr->w; i++) {
	    buffer[3*(rWidth*j+i)]   = tmpR;
	    buffer[3*(rWidth*j+i)+1] = tmpG;
	    buffer[3*(rWidth*j+i)+2] = tmpB;
	  }
      }
      gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
			 desc->header.r.x, desc->header.r.y,
			 rWidth, rHeight,
			 GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
      g_free(buffer);
      break;
    case rfbEncodingCoRRE:
      buffer = g_malloc(totPixels * 3);
      tmpR = (desc->data.corre.bgPixel >> fmt->redShift & fmt->redMax) *
	redScale;
      tmpG = (desc->data.corre.bgPixel >> fmt->greenShift & fmt->greenMax) *
	greenScale;
      tmpB = (desc->data.corre.bgPixel >> fmt->blueShift & fmt->blueMax) *
	blueScale;

      for (i = 0; i < totPixels; i++) {
	buffer[3*i] = tmpR; buffer[3*i+1] = tmpG; buffer[3*i+2] = tmpB;
      }
      for (srect = 0; srect < desc->data.corre.nSubrects; srect++) {
	guint32 pixel = desc->data.corre.rects[srect].pixel;
	rfbCoRRERectangle *sr = &(desc->data.corre.rects[srect].rect);
	tmpR = (pixel >> fmt->redShift & fmt->redMax) * redScale;
	tmpG = (pixel >> fmt->greenShift & fmt->greenMax) * greenScale;
	tmpB = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
	for (j = sr->y; j < sr->y + sr->h; j++)
	  for (i = sr->x; i < sr->x + sr->w; i++) {
	    buffer[3*(rWidth*j+i)]   = tmpR;
	    buffer[3*(rWidth*j+i)+1] = tmpG;
	    buffer[3*(rWidth*j+i)+2] = tmpB;
	  }
      }
      gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
			 desc->header.r.x, desc->header.r.y,
			 rWidth, rHeight,
			 GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
      g_free(buffer);
      break;
    case rfbEncodingHextile:
      buffer = g_malloc(totPixels * 3);
      bgR = bgG = bgB = 0;
      xPos = yPos = 0;
      for (tile = 0; tile < desc->data.hextile.nTiles; tile++) {
	rfbHextile *ht = desc->data.hextile.tiles[tile];
	
	if (ht->type & rfbHextileRaw) {
	  for (j = 0; j < ht->height; j++)
	    for (i = 0; i < ht->width; i++) {
	      guint offs = (yPos + j) * rWidth + xPos + i;
	      guint32 pixel = 0;

	      switch (bpp) {
	      case 8:  pixel = desc->data.raw[i]; break;
	      case 16: pixel = ((guint16 *)(desc->data.raw))[i]; break;
	      case 32: pixel = ((guint32 *)(desc->data.raw))[i]; break;
	      default: g_assert_not_reached();
	      }
	      buffer[offs] = (pixel >> fmt->redShift & fmt->redMax) * redScale;
	      buffer[offs+1] = (pixel >> fmt->greenShift & fmt->greenMax) * greenScale;
	      buffer[offs+2] = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
	      if (offs == 0) g_print("%06x ", pixel);
	    }
	} else {
	  if (ht->type & rfbHextileBackgroundSpecified) {
	    bgR = (ht->data.rects.bgcolour >> fmt->redShift & fmt->redMax) * redScale;
	    bgG = (ht->data.rects.bgcolour >> fmt->greenShift & fmt->greenMax) * greenScale;
	    bgB = (ht->data.rects.bgcolour >> fmt->blueShift & fmt->blueMax) * blueScale;
	    g_print("%06x ", ht->data.rects.bgcolour);
	  }
	  for (j = 0; j < ht->height; j++)
	    for (i = 0; i < ht->width; i++) {
	      guint offs = (yPos + j) * rWidth + xPos + i;
	      buffer[offs] = bgR;
	      buffer[offs+1] = bgG;
	      buffer[offs+2] = bgB;
	    }
	  if (ht->type & rfbHextileAnySubrects)
	    for (srect = 0; srect < ht->data.rects.nRects; srect++)
	      for (j = 0; j < ht->data.rects.rects[srect].h; j++)
		for (i = 0; i < ht->data.rects.rects[srect].w; i++) {
		  guint offs = (yPos + ht->data.rects.rects[srect].y + j) *
		    rWidth + xPos + ht->data.rects.rects[srect].x + i;
		  guint32 pixel = ht->data.rects.rects[srect].colour;

		  buffer[offs] = (pixel >> fmt->redShift & fmt->redMax) * redScale;
		  buffer[offs+1] = (pixel >> fmt->greenShift & fmt->greenMax) * greenScale;
		  buffer[offs+2] = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
		}
	}
	xPos += ht->width;
	if (xPos >= desc->header.r.w) {
	  xPos = 0;
	  yPos += ht->height;
	}
      }
      gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
			 desc->header.r.x, desc->header.r.y,
			 rWidth, rHeight,
			 GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
      g_free(buffer);
      break;
    default:
      g_warning("Unknown rectangle encoding returned");
    }

    gdk_draw_pixmap(GTK_WIDGET(vnc)->window,
		    vnc->gc,
		    vnc->offscreen,
		    desc->header.r.x, desc->header.r.y,
		    desc->header.r.x - vnc->hadj->value,
		    desc->header.r.y - vnc->vadj->value,
		    rWidth, rHeight);
    freeRectangleDescription(desc);
  }

  gtk_timeout_add(25, (GtkFunction)vnc_viewer_send_draw_request, vnc);
}
static void vnc_viewer_set_colourmap_entries(VncViewer *vnc,
					     rfbServerToClientMsg *msg,
					     guint16 *entries) {
  /* nothing */
}

static void vnc_viewer_bell(VncViewer *vnc, rfbServerToClientMsg *msg) {
  gdk_beep();
}
static void vnc_viewer_server_cut_text(VncViewer *vnc,
				       rfbServerToClientMsg *msg, gchar *text){
  /* I am sure I should be doing something with a selection here ... */
}

static void vnc_viewer_realize(GtkWidget *widget) {
  VncViewer *vnc = VNC_VIEWER(widget);
  GdkWindowAttr attributes;
  gint attributes_mask;

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual(widget);
  attributes.colormap = gtk_widget_get_colormap(widget);
  attributes.event_mask = gtk_widget_get_events(widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_POINTER_MOTION_MASK |
			    GDK_BUTTON_MOTION_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_KEY_PRESS_MASK |
			    GDK_KEY_RELEASE_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  if (vnc->cursor_type == VNC_VIEWER_CURSOR_DOT) {
    attributes.cursor = vnc->dot;
    attributes_mask |= GDK_WA_CURSOR;
  } else if (vnc->cursor_type == VNC_VIEWER_CURSOR_NONE) {
    attributes.cursor = vnc->none;
    attributes_mask |= GDK_WA_CURSOR;
  }

  widget->window = gdk_window_new(gtk_widget_get_parent_window(widget),
				  &attributes, attributes_mask);
  gdk_window_set_user_data(widget->window, widget);
  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  if (vnc->offscreen)
    gdk_pixmap_unref(vnc->offscreen);
  vnc->offscreen = gdk_pixmap_new(widget->window,
				  vnc->width,
				  vnc->height,
				  -1);
  if (vnc->gc)
    gdk_gc_unref(vnc->gc);
  vnc->gc = gdk_gc_new(widget->window);
  gdk_gc_copy(vnc->gc, GTK_WIDGET(vnc)->style->fg_gc[GTK_WIDGET_STATE(vnc)]);

  g_message("Sending request for full update");
  vnc_viewer_queue_event(vnc, newFramebufferUpdateRequestMsg(TRUE, 0, 0,
  						vnc->width, vnc->height));
}

static void vnc_viewer_size_allocate(GtkWidget *widget,
				     GtkAllocation *allocation) {
  VncViewer *vnc = VNC_VIEWER(widget);

  vnc->hadj->lower = 0;
  vnc->hadj->upper = vnc->width;
  vnc->hadj->step_increment = 20;
  vnc->hadj->page_increment = allocation->width;
  vnc->hadj->page_size = allocation->width;
  if (vnc->hadj->value > vnc->hadj->upper - vnc->hadj->page_size)
    vnc->hadj->value = vnc->hadj->upper - vnc->hadj->page_size;
  gtk_adjustment_changed(vnc->hadj);

  vnc->vadj->lower = 0;
  vnc->vadj->upper = vnc->height;
  vnc->vadj->step_increment = 20;
  vnc->vadj->page_increment = allocation->height;
  vnc->vadj->page_size = allocation->height;

  if (vnc->vadj->value > vnc->vadj->upper - vnc->vadj->page_size)
    vnc->vadj->value = vnc->vadj->upper - vnc->vadj->page_size;

  gtk_adjustment_changed(vnc->vadj);

  gtk_widget_queue_draw(widget);

  widget->allocation = *allocation;
  if (GTK_WIDGET_REALIZED (widget))
    gdk_window_move_resize (widget->window,
			    allocation->x, allocation->y,
			    allocation->width, allocation->height);
}

static gint vnc_viewer_expose_event(GtkWidget *widget, GdkEventExpose *event) {
  VncViewer *vnc = VNC_VIEWER(widget);

  if (GTK_WIDGET_DRAWABLE(widget)) {
    gdk_draw_pixmap(widget->window,
		    vnc->gc,
		    vnc->offscreen,
		    vnc->hadj->value + event->area.x,
		    vnc->vadj->value + event->area.y,
		    event->area.x,     event->area.y,
		    event->area.width, event->area.height);
  }
  return FALSE;
}

static gint vnc_viewer_button_press_event(GtkWidget *widget,
					  GdkEventButton *event) {
  VncViewer *vnc = VNC_VIEWER(widget);
  guint16 x, y;

  if (vnc->fd < 0) return FALSE;
  switch (event->button) {
  case 1: vnc->button_mask |= rfbButton1Mask; break;
  case 2: vnc->button_mask |= rfbButton2Mask; break;
  case 3: vnc->button_mask |= rfbButton3Mask; break;
  default:
  }

  x = vnc->hadj->value + (guint16)event->x;
  y = vnc->vadj->value + (guint16)event->y;
  vnc_viewer_queue_event(vnc, newPointerEventMsg(vnc->button_mask, x, y));
					    
  return TRUE;
}

static gint vnc_viewer_button_release_event(GtkWidget *widget,
					    GdkEventButton *event) {
  VncViewer *vnc = VNC_VIEWER(widget);
  guint16 x, y;

  if (vnc->fd < 0) return FALSE;
  switch (event->button) {
  case 1: vnc->button_mask &= ~rfbButton1Mask; break;
  case 2: vnc->button_mask &= ~rfbButton2Mask; break;
  case 3: vnc->button_mask &= ~rfbButton3Mask; break;
  default:
  }

  x = vnc->hadj->value + (guint16)event->x;
  y = vnc->vadj->value + (guint16)event->y;
  vnc_viewer_queue_event(vnc, newPointerEventMsg(vnc->button_mask, x, y));

  return TRUE;
}

static gint vnc_viewer_motion_notify_event(GtkWidget *widget,
					   GdkEventMotion *event) {
  VncViewer *vnc = VNC_VIEWER(widget);
  guint16 x, y;

  if (vnc->fd < 0) return FALSE;
  x = vnc->hadj->value + (guint16)event->x;
  y = vnc->vadj->value + (guint16)event->y;
  vnc_viewer_queue_event(vnc, newPointerEventMsg(vnc->button_mask, x, y));

  return TRUE;
}

static gint vnc_viewer_key_press_event(GtkWidget *widget,
				       GdkEventKey *event) {
  VncViewer *vnc = VNC_VIEWER(widget);

  if (vnc->fd < 0) return FALSE;
  vnc_viewer_queue_event(vnc, newKeyEventMsg(TRUE, event->keyval));

  return TRUE;
}

static gint vnc_viewer_key_release_event(GtkWidget *widget,
					 GdkEventKey *event) {
  VncViewer *vnc = VNC_VIEWER(widget);

  if (vnc->fd < 0) return FALSE;
  vnc_viewer_queue_event(vnc, newKeyEventMsg(FALSE, event->keyval));

  return TRUE;
}

static void vnc_viewer_destroy(GtkObject *object) {
  VncViewer *vnc;

  g_return_if_fail(object != NULL);
  g_return_if_fail(VNC_IS_VIEWER(object));

  vnc = VNC_VIEWER(object);

  if (vnc->hadj) gtk_object_unref(GTK_OBJECT(vnc->hadj));
  if (vnc->vadj) gtk_object_unref(GTK_OBJECT(vnc->vadj));

  vnc_viewer_close_connection(vnc);

  if (vnc->desk_name) g_free(vnc->desk_name);

  if (vnc->offscreen)
    gdk_pixmap_unref(vnc->offscreen);

  gdk_cursor_destroy(vnc->dot);
  gdk_cursor_destroy(vnc->none);
  
  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

